#!/usr/bin/env python3

import re
from pathlib import Path
from typing import Dict

from Bio import SeqIO
from Bio.Seq import Seq
from loguru import logger

from pypolca.utils.util import copy_file


def fix_consensus_from_vcf(ref_contigs: Path, vcf: Path, out_fasta: Path) -> None:
    """
    Fix errors in the consensus called in a VCF file by FreeBayes.

    Args:
        ref_contigs (Path): Path to the reference contigs in FASTA format.
        vcf (Path): Path to the VCF file containing variants.
        out_fasta (Path): Path to the output FASTA file.

    Returns:
        None
    """

    rseq = {}
    ctg, seq = "", ""
    rseq = {}

    with open(ref_contigs, "r") as file:
        for line in file:
            line = line.strip()
            if line.startswith(">"):
                if seq:
                    rseq[ctg] = seq
                ctg = line[1:]
                seq = ""
            else:
                seq += line

    if seq:
        rseq[ctg] = seq

    # Initialize variables
    ctg = ""
    fixes = []
    originals = []
    offsets = []

    total_count = 0

    # Read and process VCF file
    with open(vcf) as vcf_file:
        for line in vcf_file:
            line = line.strip()
            if line.startswith("#"):
                continue

            f = line.split()
            # this line skips the multiple allele cases
            if "," in f[4] or f[0] not in rseq:
                continue

            # if the contig is new it will instantaite a new contig - is reading line by line so needed
            if f[0] != ctg:
                if fixes:
                    if ctg not in rseq:
                        raise Exception(
                            f"sequence {ctg} not found in the input fasta file"
                        )
                    oldseq = rseq[ctg]

                    # Proceed with fixing
                    for i in range(
                        len(fixes) - 1, -1, -1
                    ):  # Going in reverse order to avoid shifting sequence due to indels
                        # First, we check if the sequence at the given offset matches the original variant
                        original_seq = oldseq[
                            offsets[i] - 1 : offsets[i] - 1 + len(originals[i])
                        ]
                        if (
                            any(c in "acgtnACGTN" for c in original_seq)
                            and not original_seq.upper() == originals[i].upper()
                        ):
                            logger.warning(
                                "WARNING! Sequence does not match the original:",
                                ctg,
                                original_seq,
                                originals[i],
                                offsets[i],
                            )
                        else:
                            # Then substitute
                            oldseq = (
                                oldseq[: offsets[i] - 1]
                                + fixes[i]
                                + oldseq[offsets[i] - 1 + len(originals[i]) :]
                            )

                    rseq[ctg] = oldseq

                fixes = []
                originals = []
                offsets = []
                ctg = f[0]

            # append if meets criteria for POLCA
            ff = f[9].split(":")
            if int(ff[5]) > 1:
                if int(ff[5]) >= 2 * int(ff[3]):
                    fixes.append(f[4])
                    originals.append(f[3])
                    offsets.append(int(f[1]))
                    total_count += 1

    # actually fix the report now
    # if fixes as previously means that this step wouldn't continue if the last contig had no changes
    # therefore, use a total count variable to achieve this.
    if total_count > 0:
        logger.info(f"POLCA has found variants. Fixing")
        # Proceed with fixing
        oldseq = rseq[ctg]
        for i in range(
            len(fixes) - 1, -1, -1
        ):  # Going in reverse order to avoid shifting sequence due to indels
            if ctg not in rseq:
                raise Exception(f"sequence {ctg} not found in the input fasta file")
            original_seq = oldseq[offsets[i] - 1 : offsets[i] - 1 + len(originals[i])]
            if (
                any(c in "acgtnACGTN" for c in original_seq)
                and not original_seq.upper() == originals[i].upper()
            ):
                logger.warning(
                    "WARNING! Sequence does not match the original:",
                    ctg,
                    original_seq,
                    originals[i],
                    offsets[i],
                )
            else:
                oldseq = (
                    oldseq[: offsets[i] - 1]
                    + fixes[i]
                    + oldseq[offsets[i] - 1 + len(originals[i]) :]
                )

        rseq[ctg] = oldseq

        records = []
        for contig, sequence in rseq.items():
            record = SeqIO.SeqRecord(seq=Seq(sequence), id=contig, description="")
            records.append(record)

        with open(out_fasta, "w") as output_file:
            SeqIO.write(records, output_file, "fasta")

    else:
        logger.warning(f"POLCA has found 0 variants.")
        logger.warning(
            f"The corrected FASTA {out_fasta} will be the same as the input FASTA {ref_contigs}."
        )
        # copy the reference to the output
        copy_file(ref_contigs, out_fasta)



import math
from pathlib import Path

from Bio import SeqIO
from loguru import logger


def create_report(vcf: Path, genome: Path, report_file: Path) -> None:
    with open(vcf, "r") as file:
        lines = file.readlines()

    # subs
    numsub = 0
    for line in lines:
        if not line.startswith("#"):
            fields = line.strip().split("\t")
            if len(fields[3]) == 1 and len(fields[4]) == 1:
                parts = fields[9].split(":")
                if int(parts[3]) == 0 and int(parts[5]) > 1:
                    # parts[3] is the number of reads supporting the ref
                    # parts[5] is the number of reads supporting the alternative
                    numsub += 1

    # index

    # Initialize variables
    nind = 0

    # Iterate through the lines and perform the desired operations
    for line in lines:
        # Skip lines starting with #
        if not line.startswith("#"):
            # Split the line using whitespace
            fields = line.split()
            # Check if the fourth and fifth fields have length greater than 1

            if len(fields[3]) > 1 or len(fields[4]) > 1:
                parts = fields[9].split(":")
                # sometimes the record field has multiple alleles (which are not changed by Polca) so int(parts[5]) returns a Valueerror.
                # in Awk it is not done as a loop but as a pipe as below. It will not count that line
                # NUMIND=`grep --text -v '^#' $BASM.vcf  |perl -ane '{if(length($F[3])>1 || length($F[4])>1){$nerr=abs(length($F[3])-length($F[4]));print "$F[9]:$nerr\n";}}' | awk -F ':' 'BEGIN{nerr=0}{if($4==0 && $6>1) nerr+=$NF}END{print nerr}'`
                # line 62 of fix_consensus_from_vcf.py deals with this in the actual polishing case
                try:
                    if int(parts[3]) == 0 and int(parts[5]) > 1:
                        nind += abs(len(fields[3]) - len(fields[4]))
                except ValueError:
                    continue

    # Print the result

    total_size = 0
    for record in SeqIO.parse(genome, "fasta"):
        total_size += len(record.seq)

    numerr = numsub + nind
    erate = numerr / total_size

    qual = round(100 - (erate) * 100, 2)

    qv = round(-10 * math.log10(erate + 0.0000000001), 2)

    with open(report_file, "w") as f:
        f.write("Stats BEFORE polishing:\n")
        logger.info("Stats BEFORE polishing:")
        f.write(f"Substitution Errors Found: {numsub}\n")
        logger.info(f"Substitution Errors Found: {numsub}")
        f.write(f"Insertion/Deletion Errors Found: {nind}\n")
        logger.info(f"Insertion/Deletion Errors Found: {nind}")
        f.write(f"Assembly Size: {total_size}\n")
        logger.info(f"Assembly Size: {total_size}")
        f.write(f"Consensus Quality Before Polishing: {qual}\n")
        logger.info(f"Consensus Quality Before Polishing: {qual}")
        f.write(f"Consensus QV Before Polishing: {qv}\n")
        logger.info(f"Consensus QV Before Polishing: {qv}")
